from ngsolve import *
import netgen.geom2d as geom2d
from netgen.geom2d import *
from math import pi
import numpy as np
import os

# Solve Helm(u):= c*u - Lap(u) = f with DG scheme
# Look for u in Vh s.t. int_Y Helm(u)*Helm(v) + int_faces (mu[grad(u)]*[grad(v)]+eta[u][v]) = int_Y f*Helm(v) for all v in Vh

hdiscretization = 1.0/(2.0*sqrt(2.0))
c = 0.0001
deg = 3
f = sin(2.0*pi*x)*sin(2.0*pi*x)*cos(2.0*pi*y)*cos(2.0*pi*y)

periodic = SplineGeometry()
pnums = [periodic.AppendPoint(*p) for p in [ (0,0), (1,0), (1,1), (0,1) ]]
lright = periodic.Append ( ["line", pnums[0], pnums[1]],bc="periodic")
btop = periodic.Append ( ["line", pnums[1], pnums[2]], bc="periodic")
periodic.Append ( ["line", pnums[3], pnums[2]], leftdomain=0, rightdomain=1, copy=lright, bc="periodic")
periodic.Append ( ["line", pnums[0], pnums[3]], leftdomain=0, rightdomain=1, copy=btop, bc="periodic")
mesh = Mesh(periodic.GenerateMesh(maxh=hdiscretization))

FEMspace = Periodic(H1(mesh, order = deg,autoupdate=True, dgjumps=True))
n = specialcf.normal(2)
h = specialcf.mesh_size

#penaltymu = 10.0*deg**2.0 
mu = 100 * deg**4 /(h**2)#10 * deg**2 / (h)

def hess(w): return w.Operator("hesse")
def Helm(w): return (c*w-hess(w)[0,0]-hess(w)[1,1])
def Lap(w): return hess(w)[0,0]+hess(w)[1,1]
u = FEMspace.TrialFunction()
v = FEMspace.TestFunction()
uh = GridFunction(FEMspace,autoupdate=True)
temph = GridFunction(FEMspace,autoupdate=True)
temph1 = GridFunction(FEMspace,autoupdate=True)
temph2 = GridFunction(FEMspace,autoupdate=True)
Rhs = LinearForm(FEMspace)
Rhs += (f * Helm(v))*dx(bonus_intorder=10)

a = BilinearForm(FEMspace, symmetric=False)	
a += (Helm(u)*Helm(v))*dx(bonus_intorder=10) + (mu*(grad(u)-grad(u).Other())*(grad(v)-grad(v).Other()))*dx(skeleton=True)

errorvals = []
residuals = []
dofs = []
def SolveProblem():
    Rhs.Assemble()
    a.Assemble()
    #uh.vec.data = a.mat.Inverse(FEMspace.FreeDofs())*Rhs.vec
    inv = a.mat.Inverse(FEMspace.FreeDofs())
    uh.vec.data = solvers.PreconditionedRichardson(a=a, rhs=Rhs.vec, pre=inv, maxit=10, tol=1e-5, printing=True)
for i in range(6):
	SolveProblem()
	temph.vec.data = a.mat*uh.vec-Rhs.vec
	temph1.vec.data = a.mat*uh.vec
	temph2.vec.data = Rhs.vec
	residuals.append(Norm(temph.vec))
	esterr =  Integrate((Helm(uh)-f)*(Helm(uh)-f),mesh)\
	+ Integrate ( (1.0/h)*(grad(uh)-(grad(uh)).Other())*(grad(uh)-(grad(uh)).Other())*dx(element_boundary=True), mesh)
	errorvals.append(sqrt(esterr))
	dofs.append(FEMspace.ndof)
	print("degrees of freedom: ",FEMspace.ndof,"Estimator error: ",sqrt(esterr))
	for el in mesh.Elements():
		mesh.SetRefinementFlag(el, True) 
	mesh.Refine()
	uh = GridFunction(FEMspace,autoupdate=True)
	temph = GridFunction(FEMspace,autoupdate=True)
	temph1 = GridFunction(FEMspace,autoupdate=True)
	temph2 = GridFunction(FEMspace,autoupdate=True)
k=0
while k < len(errorvals):
	print("Number of degrees of freedom: ",dofs[k])
	print("Linear solver residual: ",residuals[k])
	print("Error estimator:",errorvals[k])
	k = k+1
